use crate::collectors::PromMetric;
use crate::transport;
use bitflags::bitflags;
use bytes::Bytes;
use serde::Serialize;
use std::fmt::{Display, Formatter};
use std::io::Cursor;
use tokio::io::AsyncReadExt;
use tracing::{span, warn, Level};

#[derive(Serialize, Debug)]
pub struct World {
    pub id: u16,
    pub types: Vec<WorldType>,
    pub address: String,
    pub activity: String,
    pub location: WorldLocation,
    pub players: i16,
}

#[derive(Serialize, Debug)]
pub enum WorldLocation {
    Germany,
    USA,
    UK,
    Australia,
    Unknown,
}

impl From<i8> for WorldLocation {
    fn from(value: i8) -> Self {
        match value {
            0 => Self::USA,
            1 => Self::UK,
            3 => Self::Australia,
            7 => Self::Germany,
            _ => Self::Unknown,
        }
    }
}

impl Display for WorldLocation {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            WorldLocation::Germany => write!(f, "Germany"),
            WorldLocation::USA => write!(f, "USA"),
            WorldLocation::UK => write!(f, "UK"),
            WorldLocation::Australia => write!(f, "Australia"),
            WorldLocation::Unknown => write!(f, "Unknown"),
        }
    }
}

bitflags! {
    /// Bit representations of world types
    ///
    /// This is based on a best guess basis...
    #[derive(Debug, PartialEq, Eq)]
    pub struct WorldTypeRaw: i32 {
        const Members = 1;
        const PVP = 1 << 2;
        const Bounty = 1 << 5;
        const PVPArena = 1 << 6;
        const SkillTotal = 1 << 7;
        const QuestSpeedrunning = 1 << 8;
        const HighRisk = 1 << 10;
        const LastManStanding = 1 << 14;
        const SoulWars = 1 << 22;
        const Beta = 1 << 23;
        const NoSaveMode = 1 << 25;
        const Tournament = 1 << 26;
        const FreshStartWorld = 1 << 27;
        const Minigame = 1 << 28; // Not sure about this one...
        const Deadman = 1 << 29;
        const Seasonal = 1 << 30;
    }
}

#[derive(Serialize, Debug, PartialEq)]
pub enum WorldType {
    FreeToPlay,
    Members,
    PVP,
    Bounty,
    PVPArena,
    SkillTotal,
    QuestSpeedrunning,
    HighRisk,
    LastManStanding,
    NoSaveMode,
    Tournament,
    FreshStartWorld,
    Deadman,
    Beta,
    SoulWars,
    Minigame,
    Seasonal,
    Unknown,
}

impl Display for WorldType {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            &WorldType::FreeToPlay => write!(f, "FreeToPlay"),
            &WorldType::Members => write!(f, "Members"),
            &WorldType::PVP => write!(f, "PVP"),
            &WorldType::Bounty => write!(f, "Bounty"),
            &WorldType::PVPArena => write!(f, "PVPArena"),
            &WorldType::SkillTotal => write!(f, "SkillTotal"),
            &WorldType::QuestSpeedrunning => write!(f, "QuestSpeedrunning"),
            &WorldType::HighRisk => write!(f, "HighRisk"),
            &WorldType::LastManStanding => write!(f, "LastManStanding"),
            &WorldType::NoSaveMode => write!(f, "NoSaveMode"),
            &WorldType::Tournament => write!(f, "Tournament"),
            &WorldType::FreshStartWorld => write!(f, "FreshStartWorld"),
            &WorldType::Deadman => write!(f, "Deadman"),
            &WorldType::Seasonal => write!(f, "Seasonal"),
            &WorldType::Beta => write!(f, "Beta"),
            &WorldType::SoulWars => write!(f, "SoulWars"),
            &WorldType::Minigame => write!(f, "Minigame"),
            &WorldType::Unknown => write!(f, "Unknown"),
        }
    }
}

fn get_world_types(c: i32) -> Vec<WorldType> {
    let mut res: Vec<WorldType> = vec![];

    let raw = WorldTypeRaw::from_bits_retain(c);

    for world_type in raw.iter() {
        match world_type {
            WorldTypeRaw::Members => res.push(WorldType::Members),
            WorldTypeRaw::PVP => res.push(WorldType::PVP),
            WorldTypeRaw::Bounty => res.push(WorldType::Bounty),
            WorldTypeRaw::PVPArena => res.push(WorldType::PVPArena),
            WorldTypeRaw::SkillTotal => res.push(WorldType::SkillTotal),
            WorldTypeRaw::QuestSpeedrunning => res.push(WorldType::QuestSpeedrunning),
            WorldTypeRaw::HighRisk => res.push(WorldType::HighRisk),
            WorldTypeRaw::LastManStanding => res.push(WorldType::LastManStanding),
            WorldTypeRaw::NoSaveMode => res.push(WorldType::NoSaveMode),
            WorldTypeRaw::Tournament => res.push(WorldType::Tournament),
            WorldTypeRaw::FreshStartWorld => res.push(WorldType::FreshStartWorld),
            WorldTypeRaw::Beta => res.push(WorldType::Beta),
            WorldTypeRaw::Deadman => res.push(WorldType::Deadman),
            WorldTypeRaw::Seasonal => res.push(WorldType::Seasonal),
            WorldTypeRaw::Minigame => res.push(WorldType::Minigame),
            WorldTypeRaw::SoulWars => res.push(WorldType::SoulWars),
            _ => res.push(WorldType::Unknown),
        }
    }

    if res.contains(&WorldType::Unknown) {
        warn!("Unknown World Type: {:?}", raw)
    }

    if res.len() == 0 {
        res.push(WorldType::FreeToPlay)
    }

    res
}

async fn read_string(c: &mut Cursor<Bytes>) -> eyre::Result<String> {
    let mut col: Vec<u8> = vec![];

    loop {
        let item = c.read_u8().await?;
        if item == 0 {
            break;
        }

        col.push(item);
    }

    Ok(String::from_utf8(col)?)
}

async fn decode(buffer: Bytes) -> eyre::Result<Vec<World>> {
    let mut res: Vec<World> = vec![];

    let mut c = Cursor::new(buffer);

    let _buffer_size = c.read_i32().await? + 4;
    let len = c.read_i16().await?;

    for _ in 0..len {
        let id = c.read_u16().await?;
        let w = World {
            id,
            types: get_world_types(c.read_i32().await?),
            address: read_string(&mut c).await?,
            activity: read_string(&mut c).await?,
            location: WorldLocation::from(c.read_i8().await?),
            players: c.read_i16().await?,
        };

        res.push(w);
    }

    Ok(res)
}

pub async fn get() -> eyre::Result<Vec<World>> {
    let buffer = transport::http::new()
        .get("https://www.runescape.com/g=oldscape/slr.ws?order=LPWM")
        .send()
        .await
        .unwrap()
        .bytes()
        .await
        .unwrap();

    let parsing_span = span!(Level::INFO, "Parsing world data");
    let parsing_span_run = parsing_span.enter();

    let res = decode(buffer).await.unwrap();

    drop(parsing_span_run);

    Ok(res)
}

impl World {
    pub fn is_members(&self) -> bool {
        self.types.contains(&WorldType::Members)
    }
    pub fn world_type(&self) -> WorldType {
        if self.types.contains(&WorldType::QuestSpeedrunning) {
            WorldType::QuestSpeedrunning
        } else if self.types.contains(&WorldType::HighRisk) {
            WorldType::HighRisk
        } else if self.types.contains(&WorldType::LastManStanding) {
            WorldType::LastManStanding
        } else if self.types.contains(&WorldType::Bounty) {
            WorldType::Bounty
        } else if self.types.contains(&WorldType::PVP) {
            WorldType::PVP
        } else if self.types.contains(&WorldType::PVPArena) {
            WorldType::PVPArena
        } else if self.types.contains(&WorldType::NoSaveMode) {
            WorldType::NoSaveMode
        } else if self.types.contains(&WorldType::Deadman) {
            WorldType::Deadman
        } else if self.types.contains(&WorldType::Tournament) {
            WorldType::Tournament
        } else if self.types.contains(&WorldType::SkillTotal) {
            WorldType::SkillTotal
        } else if self.types.contains(&WorldType::FreshStartWorld) {
            WorldType::FreshStartWorld
        } else if self.types.contains(&WorldType::Minigame) {
            WorldType::Minigame
        } else if self.types.contains(&WorldType::SoulWars) {
            WorldType::SoulWars
        } else if self.types.contains(&WorldType::Seasonal) {
            WorldType::Seasonal
        } else if self.is_members() {
            WorldType::Members
        } else if self.types.contains(&WorldType::FreeToPlay) {
            WorldType::FreeToPlay
        } else {
            WorldType::Unknown
        }
    }
}

impl PromMetric for World {
    fn to_metric_string(&self) -> String {
        format!(
            "osrs_world_players{{id=\"{}\",location=\"{}\",isMembers=\"{:?}\",type=\"{}\"}} {}",
            self.id,
            self.location,
            self.is_members(),
            self.world_type(),
            self.players
        )
    }
}
